/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * Based on org.apache.nutch.indexer.solr.SolrDeleteDuplicates.
 * In Arch, duplicated content may exist as long as it is in different sites and/or areas.
 * For example, two sites may have overlapping sets of bookmarks with different visibility.
 * If either copy gets removed, this may lead to search failures. Finding duplicated
 * content is a smaller problem in these circumstances.
 * Only the reduce method had to be overrode, but, because some fields are declared private
 * in SolrDeleteDuplicates, have to copy and paste the whole class.
 * 
 */

package au.csiro.cass.arch.index;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.net.MalformedURLException;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.InputFormat;
import org.apache.hadoop.mapred.InputSplit;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.RecordReader;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;
import org.apache.hadoop.mapred.lib.IdentityMapper;
import org.apache.hadoop.mapred.lib.NullOutputFormat;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;
import org.apache.nutch.indexer.solr.SolrConstants;
import org.apache.nutch.indexer.solr.SolrDeleteDuplicates;
import org.apache.nutch.indexer.solr.SolrDeleteDuplicates.SolrRecord;
import org.apache.nutch.indexer.solr.SolrUtils;
import org.apache.nutch.util.NutchConfiguration;
import org.apache.nutch.util.NutchJob;
import org.apache.nutch.util.TimingUtil;
import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.SolrServer;
import org.apache.solr.client.solrj.SolrServerException;
import org.apache.solr.client.solrj.request.UpdateRequest;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.apache.solr.common.SolrDocument;
import org.apache.solr.common.SolrDocumentList;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ArchDeleteDuplicates
implements Reducer<Text, SolrDeleteDuplicates.SolrRecord, Text, SolrDeleteDuplicates.SolrRecord>,
Tool {

  public static final Logger LOG = LoggerFactory.getLogger(ArchDeleteDuplicates.class) ;

  private static final String SOLR_GET_ALL_QUERY = "*:*";

  private static final int NUM_MAX_DELETE_REQUEST = 1000;


  public static class SolrInputSplit implements InputSplit {

    private int docBegin;
    private int numDocs;

    public SolrInputSplit() { }

    public SolrInputSplit(int docBegin, int numDocs) {
      this.docBegin = docBegin;
      this.numDocs = numDocs;
    }

    public int getDocBegin() {
      return docBegin;
    }

    public int getNumDocs() {
      return numDocs;
    }

    public long getLength() throws IOException {
      return numDocs;
    }

    public String[] getLocations() throws IOException {
      return new String[] {} ;
    }

    public void readFields(DataInput in) throws IOException {
      docBegin = in.readInt();
      numDocs = in.readInt();
    }

    public void write(DataOutput out) throws IOException {
      out.writeInt(docBegin);
      out.writeInt(numDocs);
    }
  }

  public static class SolrInputFormat implements InputFormat<Text, SolrRecord> {

    /** Return each index as a split. */
    public InputSplit[] getSplits(JobConf job, int numSplits) throws IOException {
      SolrServer solr = SolrUtils.getCommonsHttpSolrServer(job);

      final SolrQuery solrQuery = new SolrQuery(SOLR_GET_ALL_QUERY);
      solrQuery.setFields(SolrConstants.ID_FIELD);
      solrQuery.setRows(1);

      QueryResponse response;
      try {
        response = solr.query(solrQuery);
      } catch (final SolrServerException e) {
        throw new IOException(e);
      }

      int numResults = 0 ;
      if ( response != null && response.getResults() != null )
                       numResults = (int)response.getResults().getNumFound();
      int numDocsPerSplit = (numResults / numSplits); 
      int currentDoc = 0;
      SolrInputSplit[] splits = new SolrInputSplit[numSplits];
      for (int i = 0; i < numSplits - 1; i++) {
        splits[i] = new SolrInputSplit(currentDoc, numDocsPerSplit);
        currentDoc += numDocsPerSplit;
      }
      splits[splits.length - 1] = new SolrInputSplit(currentDoc, numResults - currentDoc);

      return splits;
    }

    public RecordReader<Text, SolrRecord> getRecordReader(final InputSplit split,
        final JobConf job, 
        Reporter reporter)
        throws IOException {

      SolrServer solr = SolrUtils.getCommonsHttpSolrServer(job);
      SolrInputSplit solrSplit = (SolrInputSplit) split;
      final int numDocs = solrSplit.getNumDocs();
      
      SolrQuery solrQuery = new SolrQuery(SOLR_GET_ALL_QUERY);
      solrQuery.setFields(SolrConstants.ID_FIELD, SolrConstants.BOOST_FIELD,
                          SolrConstants.TIMESTAMP_FIELD,
                          SolrConstants.DIGEST_FIELD);
      solrQuery.setStart(solrSplit.getDocBegin());
      solrQuery.setRows(numDocs);

      QueryResponse response;
      try {
        response = solr.query(solrQuery);
      } catch (final SolrServerException e) {
        throw new IOException(e);
      }

      final SolrDocumentList solrDocs = response.getResults();

      return new RecordReader<Text, SolrRecord>() {

        private int currentDoc = 0;

        public void close() throws IOException { }

        public Text createKey() {
          return new Text();
        }

        public SolrRecord createValue() {
          return new SolrRecord();
        }

        public long getPos() throws IOException {
          return currentDoc;
        }

        public float getProgress() throws IOException {
          return currentDoc / (float) numDocs;
        }

        public boolean next(Text key, SolrRecord value) throws IOException {
          if (currentDoc >= numDocs) {
            return false;
          }

          SolrDocument doc = solrDocs.get(currentDoc);
          String digest = (String) doc.getFieldValue(SolrConstants.DIGEST_FIELD);
          key.set(digest);
          value.readSolrDocument(doc);

          currentDoc++;
          return true;
        }    
      };
    }
  }

  private Configuration conf;

  private SolrServer solr;

  private boolean noCommit = false;

  private int numDeletes = 0;

  private UpdateRequest updateRequest = new UpdateRequest();

  public Configuration getConf() {
    return conf;
  }

  public void setConf(Configuration conf) {
    this.conf = conf;
  }

  public void configure(JobConf job) {
    try {
      solr = SolrUtils.getCommonsHttpSolrServer(job);
      noCommit = job.getBoolean("noCommit", false);
    } catch (MalformedURLException e) {
      throw new RuntimeException(e);
    }
  }


  public void close() throws IOException {
	    try {
	      if (numDeletes > 0) {
	        LOG.info("ArchDeleteDuplicates: deleting " + numDeletes + " duplicates");
	        updateRequest.process(solr);

	        if (!noCommit) {
	          solr.commit();
	        }
	      }
	    } catch (SolrServerException e) {
	      throw new IOException(e);
	    }
	  }

  public void reduce(Text key, Iterator<SolrRecord> values, OutputCollector<Text,
          SolrRecord> output, Reporter reporter ) throws IOException
  {
    HashMap<String, SolrRecord> map = new HashMap<String, SolrRecord>() ;
    while ( values.hasNext() )
    {
      int deleted = 0 ;
      SolrRecord solrRecord = values.next() ;
      String id = solrRecord.getId() ;
      int i = id.indexOf( ':' ) ;
      if ( i > 0 ) i = id.indexOf( ':', i + 1 ) ;
      if ( i < 0 ) continue ; // strange id string, ignore
      String siteArea = id.substring( 0, i ) ;
      SolrRecord recordToKeep = (SolrRecord)map.get( siteArea ) ;
      if ( recordToKeep != null && LOG.isDebugEnabled() )
         LOG.debug( "ArchDeleteDuplicates: " + recordToKeep.getId() + " vs " + solrRecord.getId() ) ;  
      if ( recordToKeep != null && 
           ( solrRecord.getBoost() > recordToKeep.getBoost() ||
           ( solrRecord.getBoost() == recordToKeep.getBoost() && 
             solrRecord.getTstamp() > recordToKeep.getTstamp() ) ) )
      {
        updateRequest.deleteById( recordToKeep.getId() ) ; deleted = 1 ;
        recordToKeep = new SolrRecord( solrRecord ) ;
        map.put( siteArea, recordToKeep ) ;
        if ( LOG.isDebugEnabled() ) LOG.debug( "ArchDeleteDuplicates: keeping " + solrRecord.getId() ) ;  
      } else
      { 
        if ( recordToKeep != null )
        { updateRequest.deleteById( solrRecord.getId() ) ; deleted = 1 ;
          if ( LOG.isDebugEnabled() ) LOG.debug( "ArchDeleteDuplicates: keeping " + solrRecord.getId() ) ;  
        }
        else { recordToKeep = new SolrRecord( solrRecord ) ; map.put( siteArea, recordToKeep ) ; }
      }
      if ( deleted > 0 )
    	 { numDeletes++ ; reporter.incrCounter( "ArchDedupStatus", "Deleted documents", 1 ) ; }
      if ( numDeletes >= NUM_MAX_DELETE_REQUEST )
      {
        try {
              LOG.info( "ArchDeleteDuplicates: deleting " + numDeletes + " duplicates" ) ;
              updateRequest.process( solr ) ;
            } catch ( SolrServerException e ) { throw new IOException(e) ; }
        updateRequest = new UpdateRequest();
        numDeletes = 0;
      }
    }
  }

  public void dedup(String solrUrl, boolean noCommit) throws IOException {
	    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
	    long start = System.currentTimeMillis();
	    LOG.info("ArchDeleteDuplicates: starting at " + sdf.format(start));
	    LOG.info("ArchDeleteDuplicates: Solr url: " + solrUrl);
	    
	    JobConf job = new NutchJob(getConf());

	    job.set(SolrConstants.SERVER_URL, solrUrl);
	    job.setBoolean("noCommit", noCommit);
	    job.setInputFormat(SolrInputFormat.class);
	    job.setOutputFormat(NullOutputFormat.class);
	    job.setMapOutputKeyClass(Text.class);
	    job.setMapOutputValueClass(SolrRecord.class);
	    job.setMapperClass(IdentityMapper.class);
	    job.setReducerClass(ArchDeleteDuplicates.class);

	    JobClient.runJob(job);

	    long end = System.currentTimeMillis();
	    LOG.info("ArchDeleteDuplicates: finished at " + sdf.format(end) + ", elapsed: " + TimingUtil.elapsedTime(start, end));
	  }


  
  public void dedup(String solrUrl) throws IOException {
    dedup(solrUrl, false);
  }
  
  public int run(String[] args) throws IOException {
	    if (args.length < 1) {
	      System.err.println("Usage: ArchDeleteDuplicates <solr url> [-noCommit]");
	      return 1;
	    }

	    boolean noCommit = false;
	    if (args.length == 2 && args[1].equals("-noCommit")) {
	      noCommit = true;
	    }

	    dedup(args[0], noCommit);
	    return 0;
	  }

	  public static void main(String[] args) throws Exception {
	    int result = ToolRunner.run(NutchConfiguration.create(),
	        new SolrDeleteDuplicates(), args);
	    System.exit(result);
	  }

}





