/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */

package au.csiro.cass.arch.index;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import au.csiro.cass.arch.utils.ConfigList;

/**
 * Matcher of URL -> area name. Uses area filters to determine if the URL
 * belongs to an area of this site. 
 * 
 */
public class AreaMatcher
{
  Map<String, AreaFilter>      areaFilters ; // map area name -> area filter
  String                              site ; // site name                              
  
  
  /** 
   * Creates an empty area matcher.
   * @param site - site name
   */
  public AreaMatcher( String site )
  {
    areaFilters = new  HashMap<String, AreaFilter>() ;
    this.site = site ;
  }
  
  
  /** 
   * AreaMatcher factory. Creates an area matcher based on a ConfigList object.
   * @param site - site name
   */
  static public AreaMatcher newAreaMatcher( ConfigList cfg )
  {
    String[] areas = cfg.getAll( "area" ) ;
    AreaMatcher matcher = new AreaMatcher( null ) ;
    for ( int i = 0 ; i < areas.length ; i++ )
     {
       matcher.areaFilters.put( areas[ i ], AreaFilter.newAreaFilter( cfg, areas[ i ] ) ) ;
     }
    return matcher ;
  }
  
  
  /** 
   * Returns name of the area where the URL belongs, if found, else null
   * @param URL - URL to check
   * @return area name if found, else null
   */
  public String getArea( String URL )
  {
	String u = URL.toLowerCase() ;
	Iterator it = areaFilters.entrySet().iterator() ;  
	while( it.hasNext() )
	 {
	   AreaFilter filter = ((Map.Entry<String, AreaFilter>)it.next()).getValue() ;
	   if ( filter.pass( u ) ) return filter.getAreaName() ;
	 }
	return null ;
  }

  /* Incremental building functions -----------------------------------------
   * The next few functions are used to build a filter in incremental mode,
   * when roots, inclusions and exclusions can be added one by one
   */

  /** 
   * AreaMatcher factory. Creates an empty area matcher.
   * @param site - site name
   */
  static public AreaMatcher newAreaMatcher( String site )
  {
    AreaMatcher matcher = new AreaMatcher( site ) ;
    return matcher ;
  }

  /**
   * Add a root, exclusion or inclusion prefix to the filter
   * @param area - name of the area the path defines
   * @param path - root, exclusion or inclusion prefix to add
   * @param type - type of the path: 'r' - root, 'e' - exclusion, 'i' - inclusion
   */    
  public void addPath( String area, String path, String type )
  {
	AreaFilter filter = areaFilters.get( area ) ;
	if ( filter == null )
	{
	  filter = AreaFilter.newAreaFilter( area ) ;
	  areaFilters.put( area, filter ) ;
	}
	filter.addPath( path, type ) ;
  }
  
  
  /**
   * Create and add an area filter.
   * @param area - area name
   * @param roots - a list of root (seed) URLs for this area
   * @param exclusions - a list of exclusion prefixes
   * @param inclusions - a list of inclusion prefixes
   */ 
  public void addArea( String area, ArrayList<String> roots, 
		               ArrayList<String> exclusions, ArrayList<String> inclusions )
  {
    AreaFilter filter = AreaFilter.newAreaFilter( area, roots, exclusions, inclusions ) ;
    areaFilters.put( area, filter ) ;
  }

  /**
   * Finalise incremental building and make ready for filtering
   */
  public void finalise()
  {
    for ( int i = 0 ; i < areaFilters.size() ; i++ )
		  areaFilters.get( i ).finalise() ;
  }  
  //--------- End incremental building -------------------------------------------
}


