/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.index;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A "special" area for bookmarks processing. Two differences from the parent class:
 * a) This area is considered changed when at least one of the bookmark files has changed.
 * b) Bookmarks from bookmark files are used to inject URLs for processing.  
 * 
 * @author Arkadi Kosmynin
 *
 */
public class Bookmarks extends IndexArea
{
  public static final Logger LOG = LoggerFactory.getLogger( Bookmarks.class ) ;
	 
 /**
  * Default parameterless constructor, does nothing
  */
 public Bookmarks() {}
	 
	 
 /**
  * Bookmarks factory
  * 
  * @param site parent site object
  * @return Bookmarks object, a special case of IndexArea 
  * @throws Exception
  *
  */
 static public Bookmarks newBookmarks( IndexSite site ) throws Exception
 {
  Bookmarks area = new Bookmarks() ;
	   
  LOG.info( "New index area: bookmarks " ) ;
  area.name = "bookmarks" ;
  area.enabled = site.cfg.get( "enabled.bookmarks", true ) ; 
  area.site = site ;
  area.areaClause = " where site=\'" + site.name + "\' and area=\'bookmarks\'";

  area.roots = site.cfg.getAll( "file." + area.name, false ) ;
  if ( area.roots != null ) 
    System.out.println( "Listed " + area.roots.length + " bookmark files." ) ;
    
  area.setCrawlingDays() ;
  area.setExpire() ;
  area.enabled = site.cfg.get( "enabled.bookmarks", true ) ; 
  area.site.indexer.db.readDbValues( area ) ;
  
  return area ;
 }
  
 
 /**
  *  Returns true if any of the bookmark files has been changed since last area indexing.
  * 
  */ 
 @Override
 boolean changed()
 {
   if ( roots == null || roots.length == 0 ) return false ;
   Date indexed = getLastIndexed() ;
   for ( int fileNum = 0 ; fileNum <  roots.length ; fileNum++ )
   { 
	 File file = new File( roots[ fileNum ] ) ;
	 if ( !file.exists() ) continue ;
     long modTime = file.lastModified() ;
     if ( indexed.before( new Date( modTime ) ) ) return true ;
   }
   return false ;
 }
 
 /**
  *  Overrides the IndexArea injector. Takes URLs to inject from bookmark files.
  *  One URL per line.
  * 
  */
 @Override
 boolean inject() throws Exception
 {
  if ( roots == null || roots.length == 0 ) return false ;
  int lines = 0 ;
  for ( int fileNum = 0 ; fileNum <  roots.length ; fileNum++ )
   { 
	 File file = new File( roots[ fileNum ] ) ;
	 if ( !file.exists() )
	 {
	    LOG.warn( "Bookmarks file " + file.getAbsolutePath() + " does not exist" ) ;
	    continue ;
	 }
	
	 BufferedReader in = new BufferedReader( new FileReader( file ) ) ;
	 String line ;
	 
	 while( ( line = in.readLine() ) != null )
	 {  line = line.replaceAll( " ", "%20" ) ;
        site.indexer.crawlRoots.write( line + "\n" ) ;
	    lines++ ;
	 }
	 in.close() ;
   }

  if ( lines == 0 ) 
	 {
	    LOG.error( "Did not find any bookmarks. Fix file locations or switch off bookmarks area of site " + site.name ) ;
	    throw new Exception( "Did not find any bookmarks to index." ) ;
	 }
	  
  return true ;
 }

	 
	


}
