/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */

package au.csiro.cass.arch.index;

import java.text.SimpleDateFormat;
import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.csiro.cass.arch.utils.Utils;

/**
 * Implements index area - a part of a web site defined by root, include and exclude paths
 * 
 * @author Arkadi Kosmynin
 *
 */
public class IndexArea
{
 public static final Logger LOG = LoggerFactory.getLogger( IndexArea.class ) ;
 
 public static final int STATUS_NEW = 0 ;
 public static final int STATUS_FAILED = 1 ;
 public static final int STATUS_CRAWLING_DONE = 2 ;
 public static final int STATUS_MERGING_DONE = 3 ;
 public static final int STATUS_FINISHED = 9 ;
 public static String days[] = { "sun", "mon", "tue", "wed", "thu", "fri", "sat" } ;
 
 IndexSite                  site ; // reference to parent site object
 String                     name ; // name of the area
 int                    interval ; // re-indexing interval
 int                      status ; // processing status
 int                       build ; // build number
 int 					   depth ; // crawling depth, when sequential indexing
 Date                lastIndexed ; // last Date this area was indexed
 String                   expire ; // expiration query for this area
 String                  newPath ; // full path to new data directory
 String                  oldPath ; // full path to old data directory
 String[]                  roots ; // starting points for indexing this area
 String[]               oldRoots ; // starting points for indexing this area used before
 String[]          oldExclusions ; // sub-trees to exclude from indexing used before
 String[]             exclusions ; // sub-trees to exclude from indexing
 String[]          oldInclusions ; // sub-trees to include in the area used before
 String[]             inclusions ; // sub-trees to include in the area
 String[]                 doDays ; // good days for indexing this area, if any
 String                      msg ; // internal message container, for exceptions etc.
 String               areaClause ; // condition clause identifying this area in SQL statements
 boolean                 enabled ; // true - normal (default), false - disabled
 int                 buildNumber ; // Build number
 int           markedForCrawling ; // 1 if this area needs to be crawled
 boolean                  isStab ;
  
 /**
  *  Default constructor
  *
  */
 public IndexArea() {}
 
 /**
  *  Do best for releasing resources
  *
  */
 public void close() throws Exception 
 {
   name = newPath = oldPath = msg = areaClause = null ; 
   roots = oldRoots = oldExclusions = exclusions = oldInclusions = inclusions = doDays = null ;
 }
 
 /**
  *  Area factory, creates area object based on site configuration and db contents
  *  
  * @param site - IndexSite object of the site that this area is a part of 
  * @param name - area name
  * @throws Exception
  */
 static public IndexArea newIndexArea( IndexSite site, String name ) throws Exception
 {
   IndexArea area = new IndexArea() ;
   
   LOG.info( "New index area: " + name ) ;
   area.name = name ;
   area.site = site ;
   area.areaClause = " where site=\'" + site.name + "\' and area=\'" + area.name + "\'";
   
   if ( !Utils.isProperName( name ) )
      throw new Exception( "Bad name: " + name + 
                           " - must be alphanumeric starting with a letter." ) ;
   if ( name.length() > 30 )
      throw new Exception( "Too long name: " + name + " - must be up to 30 characters." ) ;
   area.exclusions = site.cfg.getAll(  "exclude." + area.name, false ) ;
   area.inclusions = site.cfg.getAll(  "include." + area.name, false ) ;
   area.roots = site.cfg.getAll( "root." + area.name, false ) ;
   if ( area.roots == null || area.roots.length == 0 )
	    throw new Exception( "No root URLs specified for area " + area.name + ", site " +
	                         site.name + "." ) ;
   area.setCrawlingDays() ;
   area.setExpire() ;
   area.enabled = site.cfg.get( "enabled." + area.name, true ) ; 
   area.site.indexer.db.readDbValues( area ) ;
   area.depth = site.cfg.getInherited( "depth." + area.name, site.depth ) ; 
      
   return area ;
 }
 
 
 /**
  * Creates all area objects for this site
  *  
  * @param site - site object
  * @return array of created area objects
  * @throws Exception
  */
 
 static public IndexArea[] newIndexAreas( IndexSite site ) throws Exception
 {
  String[] areaNames = site.cfg.getAll( "area" ) ;
  if ( areaNames == null || areaNames.length == 0 )
       throw new Exception( "No indexing areas defined.") ;
  IndexArea[] areas = new IndexArea[ areaNames.length ] ;
  for ( int i = 0 ; i < areas.length ; i++ )
      { 
        if ( areaNames[ i ].equals( "loglinks" ) ) areas[ i ] = LogLinks.newLogLinks( site ) ;
          else if ( areaNames[ i ].equals( "bookmarks" ) ) areas[ i ] = Bookmarks.newBookmarks( site ) ; 
            else areas[ i ] = newIndexArea( site, areaNames[ i ] ) ;
      }
  return areas ;
 }

 
 /**
  *  Prepares the area for crawling and does it if parallelIndexing is off.
  * 
  *  @return true if there were any URLs written to indexer's seed URL file.
  *  @throws Exception
  *  
  */
 public boolean prepare() throws Exception
 {
  if ( !needToCrawl() ) return false ;
  String areaToMark = name.equals( "loglinks" ) ? null : name ;
  String siteToMark = site.name.equals( "bookmarks" ) ? null : site.name ;
  if ( !name.equals( "bookmarks" ) && 
	   !( name.equals( "loglinks" ) && site.indexer.parallelIndexing ) )
	                   site.indexer.db.markForCrawl( siteToMark, areaToMark,  1 ) ;
  if ( !inject() ) return false ;
  // if parallel indexing is being done, there is nothing else to do
  markedForCrawling = 1 ;
  oldExclusions = exclusions ;
  oldInclusions = inclusions ;
  oldRoots = roots ;
  site.indexer.db.writeDbValues( this, true ) ;
  buildNumber = site.indexer.buildNumber ;
  setLastIndexed( site.indexer.indexingStart ) ;// in advance, as it is marked for crawl
  if ( site.indexer.parallelIndexing ) return true ;
  int res = site.indexer.crawl( site.name, name, depth ) ;
  if ( !name.equals( "bookmarks" ) ) site.indexer.db.markForCrawl( siteToMark, areaToMark,  0 ) ;
  // finalise if success
  if ( res != -1 ) finalise() ;
  return false ;
 }

  /**
  *  Finalises area crawling, saves state changes to the database in case
  *  of sequential indexing
  * 
  *  @throws Exception
  *  
  */ 
 void finalise() throws Exception
 {
   markedForCrawling = 0 ;
   site.indexer.db.writeDbValues( this, false ) ;
 }
 
 
 /**
  *  Injects URLs for crawling. Note the delta parameter. Used it to experiment with
  *  injecting very large numbers of URLs.
  * 
  */ 
 boolean inject() throws Exception
 {
  if ( roots == null || roots.length == 0 ) return false ;
  for ( int i = 0 ; i < roots.length ; i++ )
           site.indexer.crawlRoots.write( /* site.baseUrl + */ roots[ i ] + "\n" ) ;
  return true ;
 }
  
 /**
  *  Checks if this area needs to be crawled, based on the last time it had been
  *  crawled and configured crawling period. 
  *  
  *  @return true if the area has to be crawled
  */
 boolean needToCrawl()
 {
  if ( !enabled ) return false ;
  if ( markedForCrawling == 1 ) return true ; // have not finished last time
  if ( buildNumber == site.indexer.buildNumber ) return false ;
  // if list of roots or exclusions has been changed - need to recrawl
  if ( changed() ) return true ;
  
  Date nextIndexing = new Date( getLastIndexed().getTime() +
		                        (long)interval * 24 * 3600 * 1000 - 3600 * 1000 ) ;
  Date now = site.indexer.indexingStart ;
  // if there are defined preferred days and this is a repeat crawl
  if ( doDays != null && doDays.length > 1 && buildNumber > 0 )
   {
    String day = now.toString().substring( 0, 3 ) ;
    if ( Utils.in( doDays, day ) < 0 ) return false ; // a bad day to crawl
   }
  if ( nextIndexing.before( now )) return true ;
  return false ;
 }
 
 
 /**
  *  Checks if this area has changed since last crawl
  *  crawled and configured crawling period. 
  *  
  *  @return true if the area has changed
  */
 boolean changed()
 {
   if ( !Utils.isSame( roots, oldRoots ) ||
		!Utils.isSame( exclusions, oldExclusions ) || 
		!Utils.isSame( inclusions, oldInclusions ) ) return true ;
   else return false ;
 }
 
 /**
  *  Returns current date and time in text format
  */
 private static String getDate()
  {
    return new SimpleDateFormat("yyyyMMddHHmmss").format
                    ( new Date( System.currentTimeMillis() ) ) ;
  }
 
 
 void setCrawlingDays() throws Exception
 {
   String indexTimes = site.cfg.get( "interval." + name, "1" ) ;
   doDays = indexTimes.split( "[ ,]+" ) ;
   try { interval = Integer.parseInt( doDays[ 0 ] ) ;
	   } catch( Exception e )
	       {
	         throw new Exception( "First parameter must be number of days in interval, area " + name ) ;
	       }
   for ( int i = 1 ; i < doDays.length ; i++ )
	   if ( Utils.in( days, doDays[ i ] ) < 0 )
	      throw new Exception( doDays[ i ] + " is not a recognized week day in area " + name ) ;
 }
 
 void setExpire() throws Exception
 {
   SimpleDateFormat format = new SimpleDateFormat( "yyyy-MM-dd'T'kk:mm:ss.SSS'Z'" ) ;
   Date now = new Date() ;
   expire = "" ;
   
   String expire = site.cfg.get( "expire." + name, "-1" ) ;
   long aa = -1 ;
   boolean multiply = expire.charAt( 0 ) == 'x' ;
   if ( multiply ) expire = expire.substring( 1 ) ;
   try { aa = Integer.parseInt( expire ) ;
	   } catch( Exception e )
	       { throw new Exception( "Error in expire period: " + expire + " in area " + name ) ; }
   if ( multiply ) aa = aa * interval ;
   if ( aa < 0 ) return ; // never expires 
   
   now.setTime( now.getTime() - aa * 24 * 3600 * 1000 ) ;
   expire = ( "ar_site:" + site.name + " AND ar_area:" + name +
		           " AND tstamp:[* TO " + format.format( now ) + "]") ;
   site.indexer.expireQ.add( expire ) ;
 }
 
 /*
  *  Getters and setters below
  */ 
 public String getAreaClause()
  {
   return areaClause;
  }

 public void setAreaClause( String areaClause )
  {
   this.areaClause = areaClause;
  }

 public int getBuild()
  {
   return build;
  }

 public void setBuild( int build )
  {
   this.build = build;
  }

 public String[] getExclusions()
  {
   return exclusions;
  }

 public void setExclusions( String[] exclusions )
  {
   this.exclusions = exclusions;
  }

 public int getInterval()
  {
   return interval;
  }

 public void setInterval( int interval )
  {
   this.interval = interval;
  }

 public Date getLastIndexed()
  {
   if ( lastIndexed == null ) 
	   { lastIndexed = new Date() ; lastIndexed.setTime( lastIndexed.getTime() - 365 * 24 * 360 * 1000 ) ; }
   return lastIndexed ;
  }

 public void setLastIndexed( Date lastIndexed )
  {
   this.lastIndexed = lastIndexed;
  }

 public String getMsg()
  {
   return msg;
  }

 public void setMsg( String msg )
  {
   this.msg = msg;
  }

 public String getName()
  {
   return name;
  }

 public void setName( String name )
  {
   this.name = name;
  }

 public String getNewPath()
  {
   return newPath;
  }

 public void setNewPath( String newPath )
  {
   this.newPath = newPath;
  }

 public String[] getOldExclusions()
  {
   return oldExclusions;
  }

 public void setOldExclusions( String[] oldExclusions )
  {
   this.oldExclusions = oldExclusions;
  }

 public String getOldPath()
  {
   return oldPath;
  }

 public void setOldPath( String oldPath )
  {
   this.oldPath = oldPath;
  }

 public String[] getOldRoots()
  {
   return oldRoots;
  }

 public void setOldRoots( String[] oldRoots )
  {
   this.oldRoots = oldRoots;
  }

 public String[] getRoots()
  {
   return roots;
  }

 public void setRoots( String[] roots )
  {
   this.roots = roots;
  }

 public IndexSite getSite()
  {
   return site;
  }

 public void setSite( IndexSite site )
  {
   this.site = site;
  }

 public int getStatus()
  {
   return status;
  }

 public void setStatus( int status )
  {
   this.status = status;
  }

public void setOldInclusions(String[] oldInclusions) {
	this.oldInclusions = oldInclusions;
}

public String[] getOldInclusions() {
	return oldInclusions;
}

public String[] getInclusions() {
	return inclusions;
}

public void setInclusions(String[] inclusions) {
	this.inclusions = inclusions;
}

public int getBuildNumber()
 {
  return buildNumber;
 }

public void setBuildNumber( int buildNumber )
 {
  this.buildNumber = buildNumber;
 }

public boolean isStab()
 {
  return isStab;
 }



public int getMarkedForCrawling() {
	return markedForCrawling;
}



public void setMarkedForCrawling(int markedForCrawling) {
	this.markedForCrawling = markedForCrawling;
}
 
 
}
