/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.index;

import java.io.File;
import java.util.Set;

import org.apache.hadoop.conf.Configuration;
import org.apache.nutch.util.NutchConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.csiro.cass.arch.auth.Permissions;
import au.csiro.cass.arch.sql.DBInterface;
import au.csiro.cass.arch.sql.DBInterfaceFactory;
import au.csiro.cass.arch.sql.IndexSiteDB;
import au.csiro.cass.arch.utils.ConfigList;
import au.csiro.cass.arch.utils.Utils;

/**
 * Models an indexed web site, part of a multi-site index 
 * 
 * @author Arkadi Kosmynin
 *
 */
public class IndexSite
{

 public static final Logger LOG = LoggerFactory.getLogger( IndexSite.class ) ;
 public Indexer            indexer ; // root process running everything
 public ConfigList             cfg ; // site configuration parameters
 String                        url ; // base url of this site
 String                       name ; // short name/alias given to this site
 String                     dbname ; // name of used db table
 File                          dir ; // this site's directory
 File                         root ; // this site's root directory
 IndexSiteDB                    db ; // database interface
 IndexArea[]                 areas ; // site areas defined in configuration
 int                         depth ; // crawling depth 
 boolean           emailViaIndexer ; // true if use indexer object to send email
 StringBuffer                email ; // email body to send
 
 /**
  * creates a new IndexSite instance, sets dates of the first and last records
  * 
  * @param dir      directory containing site data
  * @param boss     the root Indexer process
  * @return
  * @throws Exception
  */
 static IndexSite newIndexSite( String dir, Indexer indexer ) throws Exception
 {
  File site = new File( indexer.dataDir + "/sites/" + dir ) ;
  return newIndexSite( site, indexer ) ;
 }
 
 /**
  * creates a new IndexSite instance, sets dates of the first and last records
  * 
  * @param dir      directory containing site data
  * @param boss     the root Indexer process
  * @return
  * @throws Exception
  */
 static IndexSite newIndexSite( File dir, Indexer indexer ) throws Exception
 {
  IndexSite site = new IndexSite() ;
  site.indexer = indexer ;
  site.dir = dir ;
  site.email = new StringBuffer() ;
  site.name = dir.getName() ;
   if ( !Utils.isProperName( site.name ) )
      throw new Exception( "Bad name: " + site.name + 
                           " - must be alphanumeric starting with a letter." ) ;
  site.cfg = ConfigList.newConfigList( dir.getCanonicalPath() + 
                                                     "/config.txt", indexer.cfg ) ;
  String emailRecipient = site.cfg.get( "mail.recipient", "" ) ;
  if ( emailRecipient.trim().length() == 0 ) site.emailViaIndexer = true ;
                                        else site.emailViaIndexer = false ;
  // save configuration to the database 
  indexer.db.writeConfig( site.name, site.cfg ) ;
  site.url = site.cfg.get( "URL", null ) ;
  site.depth = site.cfg.getInherited( "depth", 30 ) ;
  if ( site.url == null ) Utils.missingParam( "Base URL for site " + site.name ) ;
  if ( !site.url.endsWith("/") ) site.url += "/" ;
  // LOG.trace( "New index site: " + site.name ) ;
  Configuration nutchConf = NutchConfiguration.create() ;
  DBInterfaceFactory factory = DBInterfaceFactory.get( nutchConf ) ;
  String database = site.cfg.getInherited( "database", "MySQL" ) ;
  DBInterface intrf = factory.get( database ) ;
  site.db = intrf.newIndexSiteDB( site.cfg, site.name, false ) ;
  site.areas = IndexArea.newIndexAreas( site ) ;  
  // Clean working directories...
  String dd = dir.getCanonicalPath() + "/areas" ;
  Utils.rmdir( dd, null ) ;

  return site ;
 }

 /**
  * Does everything to release resources 
  * @throws Exception
  */
 public void close() throws Exception
 {
   indexer = null ;
   cfg = null ; 
   url = name = dbname = null ; 
   dir = root = null ;
   root = null ; 
   if ( db != null ) db.close() ; db = null ;
   if ( areas != null )
     for ( int i = 0 ; i < areas.length ; i++ )
        { areas[ i ].close() ; areas[ i ] = null ; } 
   areas = null ;
 }
 
 /**
  * Prepares all areas or a given area for crawling
  * @return true if there is something to fetch, else false 
  * @throws Exception
  */
 boolean prepare( String area, boolean firstCall ) throws Exception
 {
   String line = "site " + name ; 
   if ( area != null ) line += ", area " + area ; else line += ", all areas." ;
   System.out.println( "Preparing " + line ) ;
   appendEmail( Indexer.INFO, "Preparing " + line ) ;
   if ( firstCall )
     {
       Permissions[] permissions = cfg.getPermissions() ;
       db.applyPermissions( permissions ) ;
       db.execute( "update " + db.getTable() + " set cached=0 " ) ;
     }
   boolean result = false ;
   // if parallel indexing and call to process loglinks => at the end of processing and
   // can just mark everything unindexed for the next run and return
   if ( indexer.parallelIndexing && area != null && area.equals( "loglinks" ) )
	  db.markIndexed( false ) ;
   else 
     for ( int i = 0 ; i < areas.length ; i++ )
     {
	   String n =  areas[ i ].name ;
	    // have a particular area to process and this one is not it
       if ( area != null && !area.equals( n ) ) continue ;
        // bookmarks are processed separately in any case
       if ( area == null && "bookmarks".equals( n ) ) continue ;
    	// this is loglinks - ignore it, will be processed separately
       if ( !indexer.parallelIndexing && area == null && "loglinks".equals( n ) ) continue ;
	   result |= areas[ i ].prepare() ; // Only if time to
     }
  System.out.println( "Finished preparing " + line ) ;
  appendEmail( Indexer.INFO, "Finished preparing " + line ) ;
  sendEmail() ;
  return result ;
 }
 
 /**
  * Checks if any of the areas needs to be crawled.
  * @return true if at least one of the areas is due for crawling
  */
 boolean needToCrawl() 
 {
  for ( int i = 0 ; i < areas.length ; i++ )
     if ( areas[ i ].enabled && areas[ i ].needToCrawl() ) return true ;
  return false ;     
 }
 
 /**
  * Add a string to email message that will be sent when sendEmail is called 
  * @param level - reporting level: 0 - DEBUG, 1 - INFO, 2 - WARN, 3 - ERROR, 4 - OFF
  * @param text - text to append
  * 
  * @throws Exception
  */
 void appendEmail( int level, String text ) throws Exception
 {
   if ( emailViaIndexer )
	 indexer.appendEmail( level, text, cfg ) ;
   else Indexer.appendEmail( level, email, text, cfg ) ;
 }
 
 
 /**
  * Send email if configured
  * 
  * @throws Exception
  */
 void sendEmail( ) throws Exception
 {
   if ( !emailViaIndexer && email.length() > 0 )
	 indexer.sendEmail( email.toString(), cfg ) ;
 }
  
  
 // Getters and setters below
 
 public String getUrl()
 {
   return url;
 }


 public void setBaseUrl( String url )
 {
   this.url = url;
 }


 public Indexer getIndexer()
 {
   return indexer;
 }


 public void setIndexer( Indexer boss )
 {
   this.indexer = indexer;
 }

 public File getDir()
 {
   return dir;
 }

 public File getRoot()
 {
   return root;
 }

 public void setRoot( File root )
 {
   this.root = root;
 }


 public String getName()
  {
   return name;
  }


 public IndexSiteDB getDb()
  {
   return db;
  }


 public void setName( String name )
  {
   this.name = name;
  }
 
}
