/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.index;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Properties;

import javax.mail.Message;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.util.ToolRunner;
import org.apache.nutch.util.NutchConfiguration;
import org.apache.solr.client.solrj.SolrServer;
import org.apache.solr.client.solrj.impl.CommonsHttpSolrServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.csiro.cass.arch.logProcessing.LogProcessor;
import au.csiro.cass.arch.sql.DBInterface;
import au.csiro.cass.arch.sql.DBInterfaceFactory;
import au.csiro.cass.arch.sql.IndexRootDB;
import au.csiro.cass.arch.utils.ConfigList;
import au.csiro.cass.arch.utils.Utils;

/**
 * Main class that performs periodic indexing operations, including logs processing
 * 
 */
public class Indexer
{
 public static final Logger LOG = LoggerFactory.getLogger( Indexer.class ) ;
 static public final int DEBUG = 0, INFO = 1, WARN = 2, ERROR = 3, OFF=4 ;

 String         configPath ; // Path to configuration file used
 String          nutchHome ; // NUTCH_HOME
 ConfigList            cfg ; // Configuration parameters
 File[]              sites ; // Array of index site directories 
 IndexRootDB            db ; // Interface to a relational database
 String            dataDir ; // Data directory: root of all sites directories
 String            workDir ; // Temporary working directory to be used by Nutch
 String            solrURL ; // Solr URL to use for submission
 ArrayList<String> expireQ ; // sites expire queries 
 File                used2 ; // Used index switch
 boolean        isNewBuild ; // false if picking up interrupted build
 boolean              fork ; // fork Nutch parts in new Java instance
 boolean  parallelIndexing ; // true if to index sites in parallel
 int           buildNumber ; // last build number
 int               threads ; // parallel threads to use for crawling
 int                 depth ; // crawling depth
 int         loglinksDepth ; // loglinks crawling depth
 int                  topN ; // number of URLs to fetch at each iteration
 Configuration        conf ; // Hadoop configuration
 FileSystem             fs ; // Hadoop file system
 StringBuffer        email ;
 BufferedWriter crawlRoots ; // destination of URLs to crawl in parallel
 Date        indexingStart ;

  
  /**
   * @param args    string[] containing run parameters
   * @param args[0] (optional) site name to process
   *
   */
 public static void main( String[] args )
 {	
	 
    if ( args.length > 1 )
     { System.out.println( "Usage: Indexer [site name]" ) ; return ; }
   Indexer indexer = null ;
   try { indexer = doMain( args ) ; }
    catch( Exception e )
       {
        System.out.println( "Indexer: " + e.getMessage() ) ;
        e.printStackTrace() ;
        LOG.error( "Indexer: ", e ) ;
    	try { if ( indexer != null ) indexer.appendEmail( ERROR, e.getMessage() ) ; }
    	 catch( Exception ee ) {}
    	if ( e.getMessage().indexOf( "ock file") < 0 )
           try { lock( false ) ; } catch( Exception eee ) {} 
       }
   try {
         if ( indexer != null && indexer.email.length() > 0 )
        	          sendEmail( indexer.email.toString(), indexer.cfg ) ;
       } catch( Exception e )
       { LOG.error( "Could not send email in Indexer: " + e.getMessage() ) ; }
 }
 
 
 /**
  * @param args    string[] containing run parameters
  * @param args[0] path to configuration file to use
  * @param args[1] (optional) site name to process
  *
  */
 static Indexer doMain( String args[] ) throws Exception
 {
   ConfigList cfg ;
   IndexSite oneSite = null ;
   
   // Get configuration file from NUTCH_HOME that has to be set anyway
   String nutchHome = System.getenv( "NUTCH_HOME" ) ;
   if ( nutchHome == null )
	throw new Exception( "Stopped: environment variable NUTCH_HOME is not set." ) ;
   
   String configFile = nutchHome + "/conf/arch/config.txt" ;
   String siteToProcess = args.length == 1 ? args[ 0 ] : null ;
   lock( true ) ;
   
   Indexer indexer = new Indexer( configFile ) ;
   indexer.init() ;
   if ( siteToProcess != null )
      oneSite = IndexSite.newIndexSite( siteToProcess, indexer ) ;
   LogProcessor logProcessor = LogProcessor.newLogProcessor( configFile ) ;
   if ( oneSite != null )
   {
     logProcessor.process( "mi", siteToProcess ) ; // update list of ignored IPs
     logProcessor.process( "lw", siteToProcess ) ; // update document weights
   }
   else for ( int i = 0 ; i < indexer.sites.length ; i++ )
        {
          if ( indexer.sites[ i ] == null ) continue ;
          String name = indexer.sites[ i ].getName() ;
          logProcessor.process( i>0, "mi", name ) ; // update list of ignored IPs
          logProcessor.process( true, "lw", name ) ; // update document weights
        }
   indexer.crawl( oneSite ) ;
   indexer.expire() ;
   indexer.dedup() ;
   // now, that we have crawled everything, time to cleanUp
   indexer.cleanUp() ;
   // Here is the place to touch Nutch war file 
   indexer.appendEmail( INFO, "New index created." ) ;
   return indexer ;
 }
 
  
 /**
  * Manages crawling when parallel indexing is switched on.
  * @param oneSite - either null to crawl all sites, or the site to crawl
  * 
  */
 private void crawl( IndexSite oneSite ) throws Exception
 {
   boolean todo = false, res ;
   // For each site, prepare/crawl everything except bookmarks and log links
   if ( oneSite != null ) todo |= oneSite.prepare( null, true ) ;
   else for ( int i = 0 ; i < sites.length ; i++ )
    {
      IndexSite site = IndexSite.newIndexSite( sites[ i ], this ) ;
      todo |= site.prepare( null, true ) ;
      site.close() ;
    }
   if ( todo ) crawl( null, null, depth ) ; todo = false ;

   // For each site, prepare missed log links.
   if ( oneSite != null )todo |= oneSite.prepare( "loglinks", false ) ;
     else for ( int i = 0 ; i < sites.length ; i++ )
        {
          IndexSite site = IndexSite.newIndexSite( sites[ i ], this ) ;
	      todo |= site.prepare( "loglinks", false ) ;
          site.close() ;
        }
   // Crawl combined missed log links
   if ( todo ) crawl( "", "loglinks", loglinksDepth ) ; todo = false ;
   
   // For each site, crawl bookmarks
   expireQ.clear() ;
   if ( oneSite != null ) 
	  { if ( oneSite.prepare( "bookmarks", false ) ) crawl( oneSite.name, "bookmarks", 1 ) ; }
   else for ( int i = 0 ; i < sites.length ; i++ )
      {
        IndexSite site = IndexSite.newIndexSite( sites[ i ], this ) ;
        if ( site.prepare( "bookmarks", false ) ) crawl( site.name, "bookmarks", 1 ) ;
        site.close() ;
      }
 }

 
 /**
  * Does actual crawling. Prerequisites:
  * a) Seed URLs must be added to the crawlRoots;
  * b) All areas involved in crawl must be marked in the 'areas.status' field in the DB.  
  * 
  * @param oneSite - either null or the site to crawl
  */
 int crawl( String siteName, String areaName, int depth ) throws Exception
 {
   String aa = siteName == null ? "" : "site " + siteName ;
   if ( aa.length() > 0 ) { if ( areaName != null ) aa += ", area " + areaName ; }
	 else if ( areaName != null ) aa = " area " + areaName ;
	   else aa = "all sites and areas" ;
   LOG.info( "Crawling " + aa ) ;
   crawlRoots.close() ;
   Configuration conf = NutchConfiguration.create() ;
   String[] args = new String[ 11 ] ;
   args[ 0 ] = workDir + "/urls" ; args[ 1 ] = "-solr" ; args[ 2 ] = solrURL ;
   args[ 3 ] = "-dir" ; args[ 4 ] = workDir ;
   args[ 5 ] = "-threads" ; args[ 6 ] = "" + threads ;
   args[ 7 ] = "-depth" ; args[ 8 ] = "" + depth ;
   args[ 9 ] = "-topN" ; args[ 10 ] = "" + topN ;
   if ( areaName != null && areaName.equals( "bookmarks" ) ) 
       { args[ 8 ] = "1" ; args[ 10 ] = "-1" ;
    	 conf.set( "db.update.additions.allowed", "false" ) ;
       }
   
   if ( siteName != null ) conf.set( "arch.site", siteName ) ;
   if ( areaName != null ) conf.set( "arch.area", areaName ) ;
   conf.set( "arch.database", cfg.get( "database", "MySQL" ) ) ;
   conf.set( "arch.target.db", cfg.get( "target.db", "" )  ) ;
   conf.set( "arch.db.driver", cfg.get( "db.driver", "" ) ) ;
   conf.set( "hadoop.tmp.dir", workDir + "/tmp" ) ;
   
   int res = ToolRunner.run( conf, new ArchCrawl(), args ) ;
   if ( res < 0 )
	 throw new Exception( " Crawl failed for " + aa ) ;
   // clean and prepare for next crawl
   Utils.rmdir( workDir, null ) ; 
   Utils.mkdir( workDir ) ;
   Utils.mkdir( workDir + "/urls" ) ;
   FileWriter fstream = new FileWriter( new File ( workDir + "/urls" ).getCanonicalPath() + "/urls.txt" ) ;
   crawlRoots = new BufferedWriter( fstream ) ;
   // to get around an apparent Nutch bug
   crawlRoots.write(  "http://www.atnf.csiro.au/computing/software/arch/nutchBug.gif\n" ) ;
   db.markIndexed( buildNumber ) ;
   LOG.info( "Finished crawling " + aa ) ;   
   return res ;
 }
 
 
 /**
  * Removes expired URLs from Solr
  * 
  */
 private void expire() throws Exception
 {
   String solrURL = cfg.get( "solr.url",  "http://localhost:8080/arch" ) ;
   SolrServer server = new CommonsHttpSolrServer( solrURL ) ;
   for ( String query : expireQ )
   {
	 if ( query == null || query.length() == 0 ) continue ;
	 server.deleteByQuery( query ) ;
   }
   server.commit() ;
 }

 
 /**
  * Removes duplicated URLs from Solr
  * 
  */
 private void dedup() throws Exception
 {
  if ( cfg.get( "remove.duplicates", true ) ) 
     {
	   String solrURL = cfg.get( "solr.url",  "http://localhost:8080/arch" ) ;
       ArchDeleteDuplicates dedup = new ArchDeleteDuplicates() ;
       dedup.setConf( conf ) ;
       dedup.dedup( solrURL ) ;
     }
 }

 
 /**
  * Cleans the working and temp directories
  * 
  */
 private void cleanUp() throws Exception
 {
   Utils.rmdir( workDir, null ) ;
   Utils.mkdir( dataDir + "/done" ) ;
   String sql = "update areas set status=0, build=" + buildNumber +
                ",lastIndexed='" + DateFormat.getDateTimeInstance().format( new Date() ) + "' where status!=0" ;
   db.execute( sql ) ;
   // Update selects text
   String areaSelect = db.getAreaSelect( null ) ;
   String siteSelect = db.getSiteSelect( null ) ;
   String groupSelect = db.getGroupSelect( null, null ) ;
   if ( areaSelect != null ) saveSelect( nutchHome + "/conf/arch/areaSelect.txt", areaSelect ) ;
   if ( siteSelect != null ) saveSelect( nutchHome + "/conf/arch/siteSelect.txt", siteSelect ) ;
   if ( groupSelect != null ) saveSelect( nutchHome + "/conf/arch/groupSelect.txt", groupSelect ) ;
   lock( false ) ;
 }
 
 private void saveSelect( String fileName, String select )
 {
   try
   {
	  PrintWriter out = new PrintWriter( fileName ) ;
      out.print( select ) ;
      out.close() ;
   } catch( Exception e )
   {
	 LOG.warn( "Could not update select text file " + fileName + ", is it read/only?" ) ;  
   }
 }
  
  /**
   * Default parameterless constructor
   *
   */
 public Indexer() {} 
  
  /**
   * A normal constructor. Stores path to configuration in a member variable.
   * 
   * @param config  string path to configuration file.
   */
 public Indexer( String config )
 {
    configPath = config ;
    nutchHome = System.getenv( "NUTCH_HOME" ) ;

 }
  
  /**
   * Performs initializing based on the configuration parameters.
   * 
   * @throws Exception
   */
 public void init() throws Exception
 {
   indexingStart = new Date() ;
   email = new StringBuffer() ;
   cfg = ConfigList.newConfigList( configPath ) ;
   conf = NutchConfiguration.create();
   DBInterfaceFactory factory = DBInterfaceFactory.get( conf ) ;
   String database = cfg.get( "database", "MySQL" ) ;
   solrURL = cfg.get( "solr.url", "http://localhost:8080/arch" ) ;
   fork = cfg.get( "fork.nutch", true ) ;
   topN = cfg.get( "max.urls", 100000 ) ;
   depth = cfg.get( "depth", 30 ) ;
   loglinksDepth = cfg.get( "depth.loglinks", 1 ) ;
   threads = cfg.getInherited( "threads", 10 ) ;
   parallelIndexing = cfg.getInherited( "parallel.indexing", false ) ;
   DBInterface intrf = factory.get( database ) ;
   db = intrf.newIndexRootDB( cfg, false ) ;
    // create a list of sites
   dataDir = nutchHome + "/conf/arch" ;
   String sitesDir = dataDir + "/sites" ;
   workDir = cfg.get( "temp.dir", null ) ;
   if ( workDir == null ) workDir = dataDir + "/work" ;
   File dir = new File( sitesDir ) ;
   SiteFilter filter = new SiteFilter() ; 
   sites = dir.listFiles( filter ) ;
   expireQ = new ArrayList<String>() ;
   resetEverything() ; // Reset everything if the trigger is set  
   buildNumber = db.readInt( "select value from system where name='buildNumber'" ) ;
   if ( buildNumber == -1 ) buildNumber = 1 ;
   if ( Utils.exists( dataDir + "/done" ) ) // previous work has been finished successfully
    { 
      buildNumber++ ;
      isNewBuild = true ;
      db.execute( "delete from system where name='buildNumber'" ) ;
      db.write( "insert into system values( 'buildNumber', '" + 
                                            buildNumber + "' )", null ) ;
    } else isNewBuild = false ; // picking up an unfinished build

   if ( sites == null ) throw new Exception( "No data directory found: " + dir ) ;
   Utils.rmdir( workDir, null ) ; 
   Utils.mkdir( workDir ) ;
   String urlsDir = workDir + "/urls" ;
   Utils.mkdir( urlsDir ) ;
   conf.set( "hadoop.tmp.dir", workDir + "/tmp" ) ;
   FileWriter fstream = new FileWriter( new File ( urlsDir ).getCanonicalPath() + "/urls.txt" ) ;
   crawlRoots = new BufferedWriter( fstream ) ;
	 // to get around an apparent Nutch bug
   crawlRoots.write(  "http://www.atnf.csiro.au/computing/software/arch/nutchBug.gif\n" ) ;  
 }

 
 /**
  * Send email if configured
  * @param content - message body
  * @param cfg - configuration
  * 
  * @throws Exception
  */
 static public void sendEmail( String content, ConfigList cfg  ) throws Exception 
 {
   if ( cfg.getInherited( "mail.level", "info" ).equalsIgnoreCase( "off" ) ) return ;
   Properties props = new Properties();
   props.setProperty( "mail.transport.protocol", cfg.getInherited( "mail.transport.protocol", "smtp" ) ) ;
   if ( cfg.getInherited( "mail.host", "" ).length() == 0 ) 
	   throw new Exception( "Can't send email: mail.host is not defined." ) ;
   props.setProperty( "mail.host", cfg.getInherited( "mail.host", "" ) ) ;
   props.setProperty( "mail.user", cfg.getInherited( "mail.user", "" ) ) ;
   props.setProperty( "mail.password", cfg.getInherited( "mail.password", "" ) ) ;
   String subject = cfg.getInherited( "mail.subject", "" ) ;
   if ( subject.length() == 0 ) subject = "Arch indexing report" ;
   String addresses = cfg.getInherited( "mail.recipient", "" ) ;
   if ( addresses.length() == 0 )
	   throw new Exception( "Can't send email: mail.recipient is not defined." ) ;
   String[] to = addresses.split( "[;,:]" ) ;
   
   Session mailSession = Session.getDefaultInstance( props, null ) ;
   Transport transport = mailSession.getTransport() ;

   MimeMessage message = new MimeMessage( mailSession ) ;
   message.setSubject( subject ) ;
   message.setContent( content, "text/plain" ) ;
   for ( int i = 0 ; i < to.length ; i++ )
   { 
	 String address = to[ i ].trim() ;
	 if ( address.length() > 0 )
		message.addRecipient( Message.RecipientType.TO, new InternetAddress( address ) ) ;
   }
   if ( message.getAllRecipients() == null || message.getAllRecipients().length == 0 )
	   throw new Exception( "Can't send email: mail recipient list is invalid." ) ;
	   
   transport.connect();
   transport.sendMessage(message, message.getRecipients( Message.RecipientType.TO ) ) ;
   transport.close() ;
 }
 
 /**
  * Add a string to email message that will be sent when sendEmail is called 
  * @param level - reporting level: 0 - DEBUG, 1 - INFO, 2 - WARN, 3 - ERROR, 4 - OFF
  * @param body - email body to append to
  * @param text - text to append
  * @param cfg - configuration
  * 
  * @throws Exception
  */
 static public void appendEmail( int level, StringBuffer body, String text, ConfigList cfg )
 throws Exception
 {
   String[] codes = { "DEBUG", "INFO", "WARN", "ERROR", "OFF" } ;
   String levelLimit = cfg.getInherited( "mail.level", "INFO" ) ;
   int i ;
   for ( i = 0 ; i < codes.length ; i++ )
   { if ( codes[ i ].equalsIgnoreCase( levelLimit ) ) break ; }
   if ( i == codes.length ) throw new Exception( "Wrong mail.level parameter." ) ;
   if ( level < i ) return ;
   if ( level > codes.length - 1 ) level = codes.length - 2 ; 
   String date = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss ").format( new Date( System.currentTimeMillis() ) ) ;
   body.append( date ) ;
   body.append( codes[ level ] ) ; body.append( " " ) ;
   body.append( text ) ;
   body.append( "\r\n" ) ;
 }

 
 /**
  * Add a string to email message that will be sent when sendEmail is called 
  * @param level - reporting level: 0 - DEBUG, 1 - INFO, 2 - WARN, 3 - ERROR, 4 - OFF
  * @param text - text to append
  * 
  * @throws Exception
  */
 void appendEmail( int level, String text ) throws Exception
 {
   appendEmail( level, email, text, cfg ) ;
 }

 
 /**
  * Add a string to email message that will be sent when sendEmail is called 
  * @param level - reporting level: 0 - DEBUG, 1 - INFO, 2 - WARN, 3 - ERROR, 4 - OFF
  * @param body - email body to append to
  * @param text - text to append
  * 
  * @throws Exception
  */
 public void appendEmail( int level, String text, ConfigList cfg ) throws Exception
 {
   appendEmail( level, email, text, cfg ) ;
 }

 
 /**
  * Check, create or remove a lock file that indicates that Arch is in progress
  * @param lockIt - if true, try to create the lock. Error if it exists. If false, remove the lock.
  * 
  * @throws Exception
  */
 public static void lock( boolean lockIt ) throws Exception
 {
   String nutchHome = System.getenv( "NUTCH_HOME" ) ;
   String lockName = nutchHome + "/conf/arch/lockFile" ;
   File lockFile = new File( lockName ) ;
   if ( lockIt )
    {
     if ( lockFile.exists() )
      { throw new Exception( "Lock file " + lockName + " exists: is another copy of Arch running?" ) ; }
     try { lockFile.createNewFile() ; } catch( IOException ioe )
      { throw new Exception( "Can't create lock file " +  lockName + ", is the directory wriateable?" ) ; }
    } else
    {
      if ( lockFile.exists() && !lockFile.delete() )
    	throw new Exception( "Can't delete lock file " + lockName + 
    			                               ". Delete it befeore starting Arch again." ) ;
    }
 }
 

 /**
  * Checks for the reset trigger and resets the data if it exists. The trigger is file "reset" in the
  * configuration directory.
  * 
  * @throws Exception
  */ 
 public void resetEverything() throws Exception
 {
   File trigger = new File( dataDir + "/reset" ) ;	 
   if ( trigger.exists() )
	{ // Generally, an overkill, but sites allowed to have different databases
	  for ( File fsite : sites )
	  {
	    IndexSite site = IndexSite.newIndexSite( fsite, this ) ;
	    site.db.execute( "drop table if exists site_" + site.name ) ;
	    site.close() ;
	  }
	  db.execute( "drop table if exists areas" ) ;
	  db.execute( "drop table if exists logs" ) ;
	  db.execute( "drop table if exists roots" ) ;
	  db.execute( "drop table if exists sites" ) ;
	  db.execute( "drop table if exists system" ) ;
	  db.close();
	  DBInterfaceFactory factory = DBInterfaceFactory.get( conf ) ;
	  String database = cfg.get( "database", "MySQL" ) ;
	  DBInterface intrf = factory.get( database ) ;
	  db = intrf.newIndexRootDB( cfg, false ) ;
	  trigger.delete() ;
	  CommonsHttpSolrServer server = new CommonsHttpSolrServer( solrURL );
	  server.deleteByQuery( "*:*" ) ;
	  server.commit() ;
	}
 }

// Getters and setters 
 
public BufferedWriter getCrawlRoots() {
	return crawlRoots;
}


/**
 * A helper class, for directory filtering
 */
public static class SiteFilter implements java.io.FileFilter
{
   public boolean accept(File f)
   {
     if (f.isDirectory()) return true; else return false ;
   }
}


public String getNutchConfigPath()
{
  return nutchHome + "/conf";
}


public String getDataDir()
{
  return dataDir;
}



}
