/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.logProcessing;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.RandomAccessFile;
import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Models a log file
 * 
 * @author Arkadi Kosmynin
 *
 */
public class LogFile
{
 public static final Logger LOG = LoggerFactory.getLogger( LogFile.class ) ;

 // Stored in the database
 String          fileName ; // file name of the log file
 Date           timeStart ; // time of the first record
 Date             timeEnd ; // time of the end record
 Date        timeModified ; // file last modification date
 boolean       usedForIPs ; // true if used to make list of ignored IPs
 // Not stored in the database
 String          fullName ; // file name with full path
 String  unzippedFileName ; // without .gz
 LogSite             site ; // the site the log belongs to
 BufferedReader        br ; // input stream
 FileReader            fr ; // 
 boolean packWhenFinished ; // true if to re-pack file after finishing with it
 long                size ; // file size
 boolean           ignore ; // this log is to be ignored (a duplicate?)
  
  
 /**
 * LogFile factory. Creates a new LogFile instance, sets dates of the first
 * and last records.
 * 
 * @param fileName log file name
 * @param site  a reference to parent LogSite object
 * @return a new LogFile object
 * @throws Exception 
 *
 */
 public static LogFile newLogFile( String fileName, LogSite site ) throws Exception
 {
  LogFile logFile = new LogFile() ;
  String fname ;
  logFile.site = site ;
  logFile.fileName = fileName ;
  logFile.fullName = site.logDir + "/" + fileName ;
  LOG.info( "Log file: " + logFile.fullName ) ;

  File f = new File( logFile.fullName ) ;
  // check if this is a directory
  if ( f.isDirectory() )
  { LOG.info( "      Directory ignored: " + logFile.fullName ) ; 
    logFile.ignore = true ;
    return logFile ;
  }
  // check if this is not a known file
  if ( !site.logProcessor.db.readLogFile( logFile, site.name ) ||
       logFile.timeModified.getTime() != f.lastModified() )
  {
    LOG.info( "      New log file: " + logFile.fullName ) ;
    logFile.open() ;
    if ( !logFile.getNextGoodLine() ) logFile.ignore = true ;
         else logFile.timeStart = site.parser.getDate() ;
                    
    if ( logFile.ignore || !logFile.getLastLine() ) logFile.ignore = true ;
         else logFile.timeEnd = site.parser.getDate() ;
    logFile.setUsedForIPs( false ) ;
    logFile.close();
    if ( logFile.ignore == true )
       {
    	 LOG.info( "Ignoring log file " + logFile.fullName ) ;
         System.out.println( "      New log file " + logFile.fullName + " ignored.") ;
       } else System.out.println( "      New log file " + logFile.fullName + " from " 
    		   + logFile.timeStart + " to " + logFile.timeEnd ) ;

  }
    else
  { 
     if ( site.logProcessor.isIgnoredIPsEmpty() ) logFile.setUsedForIPs( false ) ;
     System.out.println( "      Known log file " + logFile.fullName ) ;
  }
  return logFile ;
 }
 
 /**
 * Saves db stored info of this LogFile to the database
 * 
 * @throws Exception 
 *
 */
 public void save() throws Exception
 {
   File f = new File( fullName ) ;
   timeModified = new Date( f.lastModified() ) ;
   site.logProcessor.db.saveLogFile( this, site.name ) ;
 }
  
 /**
  * Saves db stored info of this LogFile to the database
  * 
  * @throws Exception 
  *
  */
  public void delete() throws Exception
  {
    File f = new File( fullName ) ;
    f.delete() ;
  }
   
 /**
 * Opens the log file, unpacks if needed and prepares for reading
 * 
 * @throws Exception 
 *
 */
 void open() throws Exception
 {
  String fname ;
    // if the log is in zipped format (.gz), unpack it first
  if ( fullName.indexOf( ".gz" ) == fullName.length() - 3 )
   {
     Process p = Runtime.getRuntime().exec("gzip -d " + fullName ) ;
     p.waitFor() ;
     unzippedFileName = fullName.substring( 0, fullName.length() - 3 ) ;
     packWhenFinished = true ;
   } else { unzippedFileName = fullName ; packWhenFinished = false ; }
  File f = null ;
  for ( int i = 0 ; i < 1000 ; i++ )
   { f = new File( unzippedFileName ) ;
     if ( f.exists() ) break ; 
     Thread.sleep( 100 ) ; // give the OS time to refresh file tables
   }
  size = f.length() ;
  if ( !f.exists() )
   if ( packWhenFinished ) throw new Exception( " Can't unpack " + fullName ) ;
    else throw new Exception( " Log file " + fullName + " does not exist." ) ;
     
  fr = new FileReader( f ) ;
  br = new BufferedReader( fr ) ;
 }
  
 /**
 * Closes the log file, packs it back if needed 
 * 
 * @throws Exception 
 *
 */
 void close() throws Exception
 {
    br.close() ;
    fr.close() ;
    if ( packWhenFinished ) // zip this file back
      {
       File ff = new File( fileName ) ;
       if ( !ff.exists()  ) 
        { Process p = Runtime.getRuntime().exec("gzip " + unzippedFileName ) ;
          p.waitFor() ;
          for ( int i = 0 ; i < 100 ; i++ )
              { ff = new File( fileName ) ;
                if ( ff.exists() ) break ; 
                Thread.sleep( 100 ) ; // give the OS time to refresh file tables
              }
          if ( ff.exists() )
             { File ff2 = new File( unzippedFileName ) ;
               if ( ff2.exists() ) 
                        ff2.delete() ; // some gunzips do not delete the original
             }
        }
      }
 }
  
 /**
  * Returns next 'good' line that relates to a text file and not 
  * produced by a source with blacklisted IP address
  * 
  * @return LogLine object or null 
  * @throws Exception 
  *
  */
 boolean getNextGoodLine() throws Exception 
 {
   boolean result ;
   do { 
	    String ln = br.readLine() ; 
        if ( ln == null ) return false ;
        result = site.parser.parse( ln ) ;   
      } while ( !result ) ;
   return result ;
 }
   
 /**
 * Returns last line with a valid date
 * 
 * @return LogLine object or null 
 * @throws Exception 
 *
 */
 boolean getLastLine() throws Exception 
 {
  // read the last 10K of the file
  RandomAccessFile f = new RandomAccessFile( unzippedFileName, "r" ) ;
  long length = f.length() ;
  if ( length > 10000 ) length = 10000 ;
  f.seek( f.length() - length ) ;
  byte[] arr = new byte[ (int)length ] ;
  f.readFully(  arr  ) ;
  f.close() ;
  String str = new String( arr ) ;
  String[] lines = str.split( "\n" ) ;
  
  for ( int i = lines.length - 1  ; i > 0 ; i-- )
   {
     String ln = lines[ i ].trim() ;
     site.parser.parse( ln ) ;
     if ( site.parser.parseDate( ln ) ) return true ; 
   }
  // Only care for a date because there may be no 'good' line in the last few
  return false ;
 }

 // Getters and setters below
 
 public String getFileName()
  {
   return fileName;
  }

 public void setFileName( String fileName )
  {
   this.fileName = fileName;
  }

 public Date getTimeEnd()
  {
   return timeEnd;
  }

 public void setTimeEnd( Date timeEnd )
  {
   this.timeEnd = timeEnd;
  }

 public Date getTimeModified()
  {
   return timeModified;
  }

 public void setTimeModified( Date timeModified )
  {
   this.timeModified = timeModified;
  }

 public Date getTimeStart()
  {
   return timeStart;
  }

 public void setTimeStart( Date timeStart )
  {
   this.timeStart = timeStart;
  }

 public boolean isUsedForIPs()
  {
   return usedForIPs;
  }

 public void setUsedForIPs( boolean usedForIPs )
  {
   this.usedForIPs = usedForIPs;
  }
   
}


