/*
 * Takes care of Arch related filters that can be prepared by the authenticator
 * or submitted as request fields. Also, creates a query in Solr syntax from 
 * advanced query form fields.
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */


package au.csiro.cass.arch.solr;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.search.Query;
import org.apache.solr.common.SolrException;
import org.apache.solr.common.params.CommonParams;
import org.apache.solr.common.params.ModifiableSolrParams;
import org.apache.solr.common.params.ShardParams;
import org.apache.solr.common.params.SolrParams;
import org.apache.solr.common.util.StrUtils;
import org.apache.solr.handler.component.QueryComponent;
import org.apache.solr.handler.component.ResponseBuilder;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.response.SolrQueryResponse;
import org.apache.solr.search.QParser;
import org.apache.solr.search.QParserPlugin;
import org.apache.solr.search.QueryParsing;
import org.apache.solr.util.SolrPluginUtils;

import au.csiro.cass.arch.utils.Utils;

public class ArchQueryComponent extends QueryComponent
{

  /*
   * 	(non-Javadoc)
   * @see org.apache.solr.handler.component.QueryComponent#prepare(org.apache.solr.handler.component.ResponseBuilder)
   */
  @Override
  public void prepare(ResponseBuilder rb) throws IOException
  {
    SolrQueryRequest req = rb.req ;
	SolrParams params = req.getParams() ;
	HttpServletRequest request = (HttpServletRequest)req.getContext().get( "HttpServletRequest" ) ;
	// First, filter prepared by the Authenticator (if any) 
	String filter = (String)req.getContext().get( "ar_filter" ) ;
	String securityStr = (String)req.getContext().get( "ar_filter" ) ;
	// Then, extra Arch filters coming in request parameters
	if ( request != null && filter == null ) // else this is done by the Authenticator
	{
	  String sites = Utils.notEmpty( (String)request.getParameter( "ar_site" ) ) ;
	  String areas = Utils.notEmpty((String)request.getParameter( "ar_area" ) ) ;
	  String users = Utils.notEmpty((String)request.getParameter( "ar_user" ) ) ;
	  String groups = Utils.notEmpty((String)request.getParameter( "ar_groups" ) ) ;
	  filter =  Utils.makeFilter( sites, areas, users, groups ) ;
	  req.getContext().put( "ar_filter", filter ) ;
	}
    
    SolrQueryResponse rsp = rb.rsp ;

    // Set field flags
    String fl = params.get(CommonParams.FL);
    int fieldFlags = 0;
    if (fl != null) {
      fieldFlags |= SolrPluginUtils.setReturnFields(fl, rsp);
    }
    rb.setFieldFlags( fieldFlags );

    String defType = params.get(QueryParsing.DEFTYPE,QParserPlugin.DEFAULT_QTYPE);

    if (rb.getQueryString() == null)
    {
       String query = params.get( CommonParams.Q ) ;
       if ( query == null )
       { 
    	 query = Utils.makeQuery( params ) ; // could be advanced search
    	 if ( query == null || query.length() == 0 ) // still no query - send the query form
    	 {
    	   query = "ar_site:sendQueryForm" ; // this will be fast and won't find anything
           req.getContext().put( "ar_action", "sendQueryForm" ) ;
    	 }
       } else query = Utils.parseField( query, null, "", "", false ) ;
       rb.setQueryString( query ) ; 
    }

    try {
      QParser parser = QParser.getParser(rb.getQueryString(), defType, req);
      rb.setQuery( parser.getQuery() );
      rb.setSortSpec( parser.getSort(true) );
      rb.setQparser(parser);

      
      String[] fqs0 = req.getParams().getParams( CommonParams.FQ ) ;
      ArrayList<String> fqs = new ArrayList<String>() ;
      if ( fqs0 != null && fqs0.length != 0 ) Collections.addAll( fqs, fqs0 ) ;
      if ( filter != null && filter.length() > 0 ) fqs.add( filter ) ;
      if ( fqs != null && fqs.size() != 0 )
      {
        List<Query> filters = rb.getFilters() ;
        if ( filters==null ) filters = new ArrayList<Query>( fqs.size() ) ;
        for ( String fq : fqs ) 
          if ( fq != null && fq.trim().length()!=0 )
          {
            QParser fqp = QParser.getParser(fq, null, req);
            filters.add(fqp.getQuery());
          }
 
        // only set the filters if they are not empty otherwise
        // fq=&someotherParam= will trigger all docs filter for every request 
        // if filter cache is disabled
        if (!filters.isEmpty()) {
          rb.setFilters( filters );
        }
      }
    } catch (ParseException e) {
      throw new SolrException(SolrException.ErrorCode.BAD_REQUEST, e);
    }

    // TODO: temporary... this should go in a different component.
    String shards = params.get(ShardParams.SHARDS);
    if (shards != null) {
      List<String> lst = StrUtils.splitSmart(shards, ",", true);
      rb.shards = lst.toArray(new String[lst.size()]);
    }
    String shards_rows = params.get(ShardParams.SHARDS_ROWS);
    if(shards_rows != null) {
      rb.shards_rows = Integer.parseInt(shards_rows);
    }
    String shards_start = params.get(ShardParams.SHARDS_START);
    if(shards_start != null) {
      rb.shards_start = Integer.parseInt(shards_start);
    }
  }
  
  public void handleFacets( SolrQueryRequest req )
  {
	SolrParams params = req.getParams() ;
	String param = params.get( "facet" ) ;
	// If done already here, or by gateway or by Solr configuration
	String baseURL = (String)req.getContext().get( "baseURL" ) ;
	if ( param == null ) // check configuraton file and set up faceting 
	{
		
	}
	// If done by the Authenticator for a site (domain) query already
	if ( baseURL != null ) return ;
	
	baseURL = "?" + req.getParamString() ;
	baseURL = Utils.removeParam( baseURL, "ar_fc" ) ;
	String[] ar_fc = params.getParams( "ar_fc" ) ;
	if ( ar_fc != null ) // clean the base URL of the facet clause
	 {
	   String clause = Utils.makeFQ( ar_fc ) ;
	   baseURL = Utils.removeClause( baseURL, clause ) ;
	 }
	req.getContext().put( "ar_fc", ar_fc ) ;
	req.getContext().put( "baseURL", baseURL ) ;
  }

  @Override
  public String getDescription() { return "Arch QueryFilter" ; }

  @Override
  public String getSourceId() { return null ; }
	 
  @Override
  public String getSource() { return null ; }

  @Override
  public String getVersion() { return null ; }

}
