/*
 * Adds HttpServletRequest to the context to enable functions dependent on access to
 * this object in custom plugins, such as authenticators.
 * Sets up Arch default faceting based on Arch configuration files.
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */

package au.csiro.cass.arch.solr;

import javax.servlet.http.HttpServletRequest;

import org.apache.solr.common.params.ModifiableSolrParams;
import org.apache.solr.core.SolrResourceLoader;
import org.apache.solr.request.SolrQueryRequest;
import org.apache.solr.request.SolrRequestHandler;
import org.apache.solr.response.SolrQueryResponse;
import org.apache.solr.servlet.SolrDispatchFilter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.csiro.cass.arch.utils.ConfigList;
import au.csiro.cass.arch.utils.Utils;

public class DispatchFilter extends SolrDispatchFilter
{
  public static final Logger LOG = LoggerFactory.getLogger( DispatchFilter.class ) ;
	
  protected void execute( HttpServletRequest req, SolrRequestHandler handler,
                                SolrQueryRequest sreq, SolrQueryResponse rsp )
  {
    sreq.getContext().put( "HttpServletRequest", req ) ;
    String q = req.getParameter( "q" ) ;
    String fall = req.getParameter( "ar_fall" ) ;
    String ftitle = req.getParameter( "ar_ftitle" ) ;
    String furl = req.getParameter( "ar_furl" ) ;
    String fhost = req.getParameter( "ar_fhost" ) ;
    String fnot = req.getParameter( "ar_fnot" ) ;
    String fcontent = req.getParameter( "ar_fcontent" ) ;
    String facet = req.getParameter( "facet" ) ;
    try
    {
      if ( ( q != null || fall != null || ftitle != null || furl != null || fhost != null ||
    		 fnot != null || fcontent != null ) && facet == null ) // a query, but faceting is not defined on or off
      {
        String solrHome = SolrResourceLoader.locateSolrHome();
        String archDir = solrHome + "conf/arch" ;
        ConfigList rootCfg = ConfigList.newConfigList( archDir + "/config.txt" ) ;
        ConfigList cfg = rootCfg ;
        String site = req.getParameter( "ar_domain" ) ;
        if ( site != null ) cfg = ConfigList.newConfigList( archDir + "/sites/" + site + "/config.txt", rootCfg ) ;
        boolean facetSites = cfg.getInherited( "facet.sites", true ) ;
        boolean facetAreas = cfg.getInherited( "facet.areas", true ) ;
        boolean facetSizes = cfg.getInherited( "facet.sizes", false ) ;
        boolean facetDates = cfg.getInherited( "facet.dates", false ) ;
        boolean facetTypes = cfg.getInherited( "facet.types", true ) ;
        boolean faceting = cfg.getInherited( "facet", true ) ;
        faceting &= facetSites || facetAreas || facetSizes || facetDates || facetTypes ;
        
        ModifiableSolrParams params = new ModifiableSolrParams( sreq.getParams() ) ;
        if ( faceting ) params.add( "facet", "on" ) ; else params.add( "facet", "on" ) ;
        if ( facetSites && faceting ) params.add( "facet.field", "ar_site" ) ;
        if ( facetAreas && faceting ) params.add( "facet.field", "ar_area" ) ;
        if ( facetTypes && faceting ) params.add( "facet.field", "type" ) ;
        if ( facetSizes && faceting ) // could be done with "standard" facet.range params, but want non-linear scale
        {
            params.add( "facet.query", "contentLength:[0 TO 1023]" ) ;  // < 1K
            params.add( "facet.query", "contentLength:[1024 TO 10239]" ) ;  // < 10K
            params.add( "facet.query", "contentLength:[10240 TO 51199]" ) ; // < 50K
            params.add( "facet.query", "contentLength:[51200 TO 102399]" ) ; // < 100K
            params.add( "facet.query", "contentLength:[102400 TO 511999]" ) ; // < 500K
            params.add( "facet.query", "contentLength:[512000 TO 1048575]" ) ; // < 1M
            params.add( "facet.query", "contentLength:[1048576 TO 5242879]" ) ; // < 5M
            params.add( "facet.query", "contentLength:[5242880 TO 10485759]" ) ; // < 10M
            params.add( "facet.query", "contentLength:[10485760 TO 52428799]" ) ; // < 50M
            params.add( "facet.query", "contentLength:[52428800 TO 104857599]" ) ; // < 100M
            params.add( "facet.query", "contentLength:[104857600 TO 524287999]" ) ; // < 500M
            params.add( "facet.query", "contentLength:[524288000 TO *]" ) ; // >= 500M
        }
        if ( facetDates && faceting ) // could be done with "standard" facet.date params, but want non-linear scale
        {
            params.add( "facet.query", "lastModified:[NOW/DAY-7DAYS TO NOW/DAY+1DAY]" ) ;  
            params.add( "facet.query", "lastModified:[NOW/DAY-1MONTH TO NOW/DAY-7DAYS]" ) ; 
            params.add( "facet.query", "lastModified:[NOW/DAY-3MONTHS TO NOW/DAY-1MONTH]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-6MONTHS TO NOW/DAY-3MONTHS]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-1YEAR TO NOW/DAY-6MONTHS]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-2YEARS TO NOW/DAY-1YEAR]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-3YEARS TO NOW/DAY-2YEARS]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-4YEARS TO NOW/DAY-3YEARS]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-5YEARS TO NOW/DAY-4YEARS]" ) ;
            params.add( "facet.query", "lastModified:[NOW/DAY-10YEARS TO NOW/DAY-5YEARS]" ) ;
            params.add( "facet.query", "lastModified:[* TO NOW/DAY-10YEARS]" ) ;
        }
        sreq.setParams( params ) ;
      }	
    } catch( Exception e )
    { LOG.error( "Could not set up faceting: " + e.getMessage() ) ; }
    super.execute( req, handler, sreq, rsp ) ;
  }
	

}
