/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.sql;

import java.util.ArrayList;
import java.util.Date;
import java.util.Map;

import au.csiro.cass.arch.index.IndexArea;
import au.csiro.cass.arch.logProcessing.LogFile;
import au.csiro.cass.arch.utils.ConfigList;

/**
 * A set of db functions concerning tables shared by all sites and areas
 * 
 * @author Arkadi Kosmynin
 *
 */
public interface IndexRootDB extends ServletDB, IndexDB
{

 /**
 * Register (a new) site
 * 
 * @param site  name of the site
 * @param baseURL base url of the site, e.g. http://www.mysite.com
 */
 public void regSite( String site, String baseURL ) throws Exception ;

 /**
 * Saves site configuration to the db
 * 
 * @param site  name of the site
 * @param cfg configuration object
 */
 public void writeConfig( String site, ConfigList cfg ) throws Exception ;

 /**
 * Reads site configuration from the db
 * 
 * @param site  name of the site
 * @return site configuration object
 */
 public ConfigList readConfig( String site ) throws Exception ;

 /**
 * Get date of a log processing point 
 * 
 * @param point   log processing point
 * @param site    site name
 * @return Date of the log processing point
 */
 public Date getProcessed( String point, String site ) throws Exception ;

 /**
 * Write date of a log processing point 
 * 
 * @param point   log processing point
 * @param site    site name
 * @param value   date to write
 */
 public void putProcessed( String point, String site, Date value )
                                                       throws Exception ;

 /**
 * Saves log file info 
 * 
 * @param logFile logFile object
 * @param site    site name
 */
 public void saveLogFile( LogFile logFile, String site ) throws Exception ;

 /**
 * Reads log file info 
 * 
 * @param logFile logFile object, must contain log file name
 * @param site    site name
 * @return  true if log file was found, else false
 */
 public boolean readLogFile( LogFile logFile, String site ) throws Exception ;

 /**
 * Deletes info on all log files for a given site
 * 
 * @param site    site name
 */
 public void deleteLogFiles( String site ) throws Exception ;

 
 /**
 * Deletes info on log processing points for a given site
 * 
 * @param site    site name
 */
 public void resetWeights( String site ) throws Exception ;
 
 
 /**
 * For a given site, for each log file, deletes info on whether the log file
 * was used to construct the list of blocked IPs
 * 
 * @param site    site name
 */
 public void resetIPs( String site ) throws Exception ;
 
 
 /**
 * Drops all central tables 
 */
 public void resetAll() throws Exception ;
 
  /**
  *  Closes database connection if it is open.
  */
 public void close() throws Exception ;
 
  
 /**
  *  Constructs a html select element for site selection.
  *  
  *  @param selected   name of selected site that will be marked as SELECTED
  *  @return    text of select element ready to insert in a html file
  */
 public String getSiteSelect( String selected ) throws Exception ;
  

 /**
  *  Constructs a html select element for area selection.
  *  
  *  @param selected   name of selected area that will be marked as SELECTED
  *  @return    text of select element ready to insert in a html file
  */
 public String getAreaSelect( String selected ) throws Exception ;
  
 
 /**
  *  Constructs a group select html select element where areas are groupped by site.
  *  
  *  @param selectedSite   name of site containing selected area, or null
  *  @param selectedArea   name if selected area, if selectedSite is not null
  *  @return    text of group select element ready to insert in a html file
  */
 public String getGroupSelect( String selectedSite, String selectedArea ) throws Exception ;
 
 /**
  * Read db stored information on this area
  * 
  * @param area IndexArea object, must have name and site set
  */
  public void readDbValues( IndexArea area ) throws Exception ;

  /**
  * Write db stored information on this area
  * 
  * @param area IndexArea object
  * @param arraysToo    if true, then db stored arrays will be saved too
  */
  public void writeDbValues( IndexArea area, boolean arraysToo ) throws Exception ;
  
  /**
   * Marks given area for crawling. This mark is used in filters.
   *
   * @param site site name
   * @param area area name
   * @param value 0 - not to crawl, 1 - to crawl
   * 
   * @throws Exception
   */ 
  public void markForCrawl( String site, String area, int value ) throws Exception ;
  
  /**
   * Marks areas marked for crawling as indexed. Sets indexing time and build number.
   *
   * @param buildNumber - latest build number
   * 
   * @throws Exception
   */ 
  public void markIndexed( int buildNumber ) throws Exception ;
  
  
  /**
   * Reads roots, includes and excludes of marked areas to a Map
   *
   * @return a map of root, include and exclude paths for all areas marked for crawling
   * @throws Exception
   */ 
  public Map readRoots() throws Exception ;

}
