/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.sql;

import java.io.BufferedReader;
import java.io.OutputStream;
import java.sql.ResultSet;

import au.csiro.cass.arch.auth.Permissions;
import au.csiro.cass.arch.index.IndexArea;
import au.csiro.cass.arch.index.LogLinks;
import au.csiro.cass.arch.logProcessing.ScoreFile;

/**
 * A set of db related functions needed by IndexSite and LogSite objects
 * 
 * @author Arkadi Kosmynin
 *
 */
public interface IndexSiteDB extends ServletDB, IndexDB
{
  
  /**
   * Read <url, weight> pairs and export them to a stream in SiteMap xml format
   * 
   * @param file  output file
   */

  void exportSiteMap( OutputStream o, String baseURL, float norm ) throws Exception ;

  /**
   * Read <url, weight> pairs from a stream and save them to the db
   * 
   * @param file  output file
   */
  void importSiteMap( BufferedReader rd, String baseURL, float norm ) throws Exception ;
 
  /**
   * Read <url, score> pairs and export them to a file
   * 
   * @param file  output file
   */
	
  void scores2file( ScoreFile file ) throws Exception ;

  /**
   * Read <url, score> pairs from a file and save them to the db
   * 
   * @param file  output file
   */
  void file2scores( ScoreFile file ) throws Exception ;
	   
  /**
   * Init reading document scores from the db
   * 
   * @return ResultSet object to obtain scores from 
   */

  ResultSet listScores() throws Exception ;
  
  /**
   * Obtain max document weight value from the db
   * 
   * @return max weight 
   */
  float getMaxWeight() throws Exception ;
  
  /**
  * Obtain max document score value from the db
  * 
  * @return max score 
  */
  int getMaxScore() throws Exception ;
  
  /**
  * Reset document scores
  * 
  */
  void resetScores() throws Exception ;

  /**
  * Update weight of a document 
  * 
  * @param id   document id in the database
  * @param weight   weight value to set
  */
  void updateWeight( int id, float weight ) throws Exception ;
  
  /**
  * Multiply scores in the db by a given koefficient
  * 
  * @param koeff    koefficient
  */
  void normaliseWeights(  float koeff ) throws Exception ;
    
  /**
  * Update or insert a document score
  * 
  * @param path path to the document
  * @param name document name
  * @param score    document score
  * @param insert if true, a record will be created if no such document
  */
  void writeScore( String path, String name, float score, boolean insert ) throws Exception ;
    
  
  /**
  * Read weight of a document
  * 
  * @param id db id of the document
  * @return document weight
  */  
  float getWeight( int id ) throws Exception ;

  /**
  * Read weight of a document
  * 
  * @param url url of the document
  * @return document weight
  */  
  float getWeight( String url ) throws Exception ;
  
  /**
  * Create a document record + all ancestors, if it does not exist
  * 
  * @param url url of the document
  * @param markIndexed if true, mark this document as indexed
  * @return document IndexNode
  */    
  IndexNode makeSureExists( String url, boolean markIndexed ) throws Exception ;

  /**
   * Apply permissions set via configuration file
   * 
   * @param permissions array of Permissions objects
   */    
  void applyPermissions( Permissions[] permissions ) throws Exception ;

  /**
   * Read urls existing in the db, not found by the crawler, to injector input file
   * 
   * @param area area for which to do this
   * @return int count of output URLs
   */    
  int readUnindexedURLs( LogLinks area ) throws Exception ;

  /**
   * Mark all site URLs as indexed or not, depending on the parameter
   * 
   * @param boolean indexed - mark as indexed if true
   */     
  public void markIndexed( boolean indexed ) throws Exception ;
  
  /**
  * Read arrays of inclusions, exclusions and roots for this area
  * 
  * @param area area for which to do this
  */    
  void readParams( IndexArea area ) throws Exception ;

  /**
  * Write inclusions, exclusions and roots of this area to the db
  * 
  * @param area area for which to do this
  */      
  void writeParams( IndexArea area ) throws Exception ;
    
  /**
  * Close db connection, release resources
  */      
  void close() throws Exception ;

  /**
  * Connect to the db, init statements
  */      
  void connect() throws Exception ;
  
  /**
  * Reset document weights
  */      
  void resetWeights() throws Exception ;
  
  /**
  * Drop site db table
  */        
  void resetAll() throws Exception ;

// These are needed by IndexRootDB, so, have to publish them  
  
  /**
   * Read document node info from the database, re-connect if needed
   * The document is identified by either the id, or path and name. 
   * 
   * @param id id of the document
   * @param path path to the document
   * @param name name of the document
   * @return document id
   */   
  IndexNode getNode( int id, String path, String name ) throws Exception ;
  
  /**
   * Write/update document node info, re-connect if needed
   * 
   * @return document id
   */   
  public int updateNode( IndexNode n ) throws Exception ;
  
  /**
   * @return the site table name
   */
  public String getTable() ;


}