/*
 * 
 * Copyright (c) CSIRO Astronomy and Space Science
 * Author Arkadi.Kosmynin@csiro.au
 * 
 */
package au.csiro.cass.arch.logProcessing;

import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.StringTokenizer;

import org.apache.hadoop.conf.Configuration;
import org.apache.nutch.net.URLNormalizers;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import au.csiro.cass.arch.utils.Utils;

/**
 * Log line parser for combined log format
 * 
 */
public class LogLineParserCombined implements LogLineParser
{
 public static final Logger LOG = LoggerFactory.getLogger( LogLineParserCombined.class ) ;
 // extensions of common web pages
 static String[] logPages = new String[] { ".htm", ".html", ".php", ".php3", ".php4", ".php5", ".jsp", ".do", 
	                                       ".txt", ".pdf", ".asp", ".aspx" } ; 

 private Date                    date ; // date of the log line
 private LinkedList              addr ; // client IP addresses
 private String                   url ; // requested URL
 private String          canonicalUrl ; // canonical URL 
 private int                   result ; // response code
 private SimpleDateFormat   formatter ; // date formatter
 private                 boolean good ; // true if the line is good
 private LogProcessor    logProcessor ; // environment
 private LogSite              logSite ; // site in processing
 String                          line ; // line submitted for parsing
 String[]                   textPages ; // a list of text pages extensions
 String[]                 ignoreFiles ; // a list of files to ignore in log processing, like images
 URLNormalizers           normalizers ;
 
 Configuration                  conf ;

 public LogLineParserCombined()
 { 
   addr = new LinkedList() ;
   this.formatter = new SimpleDateFormat( "dd/MMM/yyyy:HH:mm:ss Z" ) ;
   good = false ;
 }
  
 /**
  * Parse a log line 
  * 
  * @param ln - input string. Normally, a line read from a log file.
  * @return true if a valid line was parsed, false if the line should be ignored.
  */
  public boolean parse( String ln )
  {
	return parse( ln, false ) ;  
  }

  /**
   * Extract date from a log line 
   * 
   * @param ln input string. Normally, a line read from a log file.
   * @return true if a date was found and parsed, else false.
   */
   public boolean parseDate( String ln )
   {
 	return parse( ln, true ) ;  
   }
  /**
   * Parse a log line 
   * 
   * @param ln - input string. Normally, a line read from a log file.
   * @param forDate - true if only extract date from line, false if do full parsing and checking 
   */
  public boolean parse( String ln, boolean forDate )
  {
   line = ln ;
   try 
   {
    addr.clear() ;
    good = false ;
    date = null ;
    boolean badIP = false ;
    int pos ;
  
    pos = ln.indexOf( '?' ) ;
    if ( pos >= 0 )
       ln = ln.substring( 0, pos ) ; 
    // split the line into tokens
    StringTokenizer tn = new StringTokenizer( ln, " " ) ;
    // get all addresses
    int i = 0 ;
    do { 
         String a = tn.nextToken() ;
         i = a.indexOf( ',' ) ;
         if ( i > 0 ) a = a.substring( 0, i ) ; 
         if ( a.equals( "-" ) ) badIP = true ; // will return a bad line that will be ignored 
         addr.add( a ) ;
       } while( i > 0 ) ; // while comma separated ip addresses

    // This can be done faster because most lines have only one address
    if ( !forDate && !validIP() ) return false ; // ignore line with this IP 

    tn.nextToken() ; tn.nextToken() ;
    String dateString1 = tn.nextToken().substring( 1 ) ;
    String dateString2 = tn.nextToken() ;
    i = dateString1.indexOf( ']' ) ;
    if ( i < 0 )
      { dateString1 = dateString1 + " " + dateString2.substring( 0, dateString2.length() - 1 ) ;
        tn.nextToken() ; // skip "GET 
      }
        else dateString1 = dateString1.substring( 0, dateString1.length() - 1 ) ;
    try { 
          date = formatter.parse( dateString1 ) ;
        } catch( Exception e ) { return false ; } // must be a bad date
    if ( forDate ) return true ; // date has been found, nothing else matters     
    if ( logProcessor.filterByAgent && !validAgent( ln ) ) // must be a search engine
                          return false ;
    url = logSite.getUrl() + tn.nextToken().substring( 1 ) ;
    if ( normalizers != null ) url = normalizers.normalize( url, "logs" ) ;
    if ( !validURL( url ) )
                         return false ;
    URL u = new URL( url ) ;
    url = u.getPath() ;
    tn.nextToken() ; // skip HTTP/1.1"
    String code = tn.nextToken() ;
    result = Integer.parseInt( code ) ;
    // consider only served pages, ignore redirections and errors
    if ( result != 200 && result != 304 ) return false ;
    
//    tn.nextToken() ;
//    String ref = tn.nextToken() ;  
//    if ( !cont.validURL( ref ) ) // is referer a valid URL?
//                         return cont ;

    if ( badIP || !validIP() )
                        return false ;
    
    good = true ;
    // System.out.println( dateString1 + " " + cont.url ) ;
  } catch( Exception e )
   {
	 if ( LOG.isDebugEnabled() )
	 { 
	   String s = "Error parsing line: " + ln ;
	   if ( e.getMessage() != null ) s += " " + e.getMessage() ;
       LOG.debug( s ) ;
	 }  
   }
  
  return good ; 
 }

 /**
 * Check if the line is generated by an accepted user agent. Opera, MSIE
 * and Firefox are more than enough for statistical logs processing.
 * 
 * @param ln user agent substring of the log line being parsed
 * @return true is the agent is good
 */
 static public boolean validAgent( String ln )
  {
   int i = ln.indexOf( "\" \"" ) ;
   if ( i > 0 )
    {
     String agent = ln.substring( i + 3 ).toLowerCase() ;
     if ( agent.indexOf( "opera/" ) >= 0 ) return true ;
     if ( agent.indexOf( "mozilla/" ) >= 0 ) 
        if ( agent.indexOf( "msie" ) > 0 || agent.indexOf( "firefox" ) > 0  )
                                           return true ;
    }
   return false ;
  }
  
 /**
 * Check if the line is not generated by a source with blacklisted IP address.
 * 
 * @return true if the IP address is not blacklisted
 */
 public boolean validIP()
  {
    if ( logProcessor.isIgnoredIPsEmpty()  ) return true ; // accept all IPs 
    // try to determine if these addresses are to be ignored
    int i = 0 ;
    while( i < addr.size() )
     { String a = (String)addr.get( i ) ;
       if ( a.equals( "-" ) || logProcessor.getIgnoredIPs().contains( a ) )
                           addr.remove( i ) ;
          else i++ ;
     }
    if ( addr.size() == 0 )
                       return false ; // bad, all addresses are ignored 
    return true ;
  }
 
 /**
 * Filter URLs of common picture files and bogus URLs generated
 * by hackers
 * 
 * @return true if the path refers to a parseable file
 */
 public boolean validURL( String path )
  {
	// cut off arguments, if any
	path = path.toLowerCase() ;
	int q = path.indexOf( '?' ) ;
	String p = path ;
	if ( q > 0 ) p = path.substring( 0, q ) ;
    int dot = p.indexOf( '.' ), is, j ;
    if ( dot >= 0 )
    {
      for ( j = 0 ; j < ignoreFiles.length ; j++ )
    	 if ( (is = p.indexOf( ignoreFiles[ j ], dot ) ) >= 0 // image request
    			 && is + ignoreFiles[ j ].length() == path.length() ) return false ;
      
      for ( j = 0, is = -1 ; j < logPages.length ; j++ )
          if ( (is = p.indexOf( logPages[ j ], dot ) ) >= 0 )
          {
             int i1 = p.indexOf( '.', is + 2 ) ;
             int i2 = p.indexOf( '/', is + 2 ) ;
             if ( i1 >= 0 || i2 >= 0 )
             { // hackers
               String a = (String)addr.get( 0 ) ;	
               logProcessor.getIgnoredIPs().add( a ) ;     
               return false ;
             }
          }
    }
    
    if ( path.indexOf( "?n=http://" ) > 0 ) // hackers
    {  
      String a = (String)addr.get( 0 ) ;	
      logProcessor.getIgnoredIPs().add( a ) ;     
      return false ;
    }
    if ( path.equals( "/robots.txt" ) ) 
                                       return false ;
    return true ;
  }
  
 // Getters and setters below
 
 public LinkedList getAddr()
  {
   return addr;
  }

 public Date getDate()
  {
   return date;
  }

 public int getCode()
  {
   return result;
  }

 public String getUrl()
 {
  return url;
 }

 public String getCanonicalUrl()
 {
  if ( canonicalUrl == null ) return url ;
     else return canonicalUrl;
 }

 public boolean isGood()
  {
   return good ;
  }

 public Configuration getConf()
  {
   return conf;
  }

 public void setConf( Configuration conf )
  {
    this.conf = conf ;
    normalizers = new URLNormalizers( conf, "logs" ) ;
  }
 
 public LogSite getLogSite()
  {
   return logSite ;
  }

 public void setLogSite( LogSite logSite )
  {
   this.logSite = logSite;
   this.logProcessor = logSite.getBoss() ;
   textPages = logProcessor.getTextPages() ;
   ignoreFiles = logSite.getIgnoreFiles() ;
  }

}
